import re
from datetime import date, datetime


class SpaydGenerator:
    """SPAYD (Short Payment Descriptor) string generator."""

    RE_ACCOUNT = re.compile(r"((?P<ba>\d+(?=-))-)?(?P<a>\d+)/(?P<b>\d{4})")

    def __init__(
        self,
        account,
        bic=None,
        amount=None,
        currency=None,
        x_vs=None,
        x_ss=None,
        x_ks=None,
        alternate_accounts=None,
        recipient_name=None,
        due_date=None,
        payment_type=None,
        message=None,
        notification_type=None,
        notification_address=None,
        x_per=None,
        x_id=None,
        x_url=None,
        reference=None,
    ):
        """
        http://qr-platba.cz/pro-vyvojare/specifikace-formatu/

        :param account: ACC account number, can be specified either as IBAN or in CZ format 12-123456789/0300
        :param bic: BIC/SWIFT code (optional, appended to ACC as IBAN+BIC)
        :param amount: AM payment amount
        :param currency: CC currency (3 letters)
        :param x_vs: X-VS
        :param x_ss: X-SS
        :param x_ks: X-KS
        :param alternate_accounts: ALT-ACC
        :param recipient_name: RN recipient name
        :param due_date: DT due date, IS0 8601
        :param payment_type: PT max 3 digits
        :param message: MSG message for recipient, max 60 chars
        :param notification_type: NT P for phone or E for email
        :param notification_address: NTA
        :param x_per: X-PER number of days to repeat payment if unsuccessful
        :param x_id: X-ID
        :param x_url: X-URL
        :param reference: RF recipient reference number. Max 16 digits. integer.
        """
        if account is None:
            raise ValueError("account is required")
        self.account = account
        self.bic = bic
        self.amount = amount
        self.currency = currency
        self.x_vs = x_vs
        self.x_ss = x_ss
        self.x_ks = x_ks
        self.alternate_accounts = alternate_accounts
        self.recipient_name = recipient_name
        self.due_date = due_date
        self.payment_type = payment_type
        self.message = message
        self.notification_type = notification_type
        self.notification_address = notification_address
        self.x_per = x_per
        self.x_id = x_id
        self.x_url = x_url
        self.reference = reference

    def _convert_to_iban(self, match):
        """Convert czech account number to IBAN from a RE_ACCOUNT match object."""
        iban = "CZ00{b}{ba:0>6}{a:0>10}".format(
            ba=match.group("ba") or 0,
            a=match.group("a"),
            b=match.group("b"),
        )

        # convert IBAN letters into numbers
        crc = re.sub(r"[A-Z]", lambda m: str(ord(m.group(0)) - 55), iban[4:] + iban[:4])

        # compute control digits
        digits = f"{98 - int(crc) % 97:0>2}"

        return iban[:2] + digits + iban[4:]

    @property
    def _account(self):
        if self.account is not None:
            m = self.RE_ACCOUNT.match(self.account)
            if m:
                iban = self._convert_to_iban(m)
            else:
                iban = self.account
            if self.bic:
                return f"ACC:{iban}+{self.bic}*"
            return f"ACC:{iban}*"
        return ""

    @property
    def _alternate_accounts(self):
        if self.alternate_accounts:
            formatted = []
            for account in self.alternate_accounts:
                m = self.RE_ACCOUNT.match(account)
                if m:
                    formatted.append(self._convert_to_iban(m))
                else:
                    formatted.append(account)
            return "ALT-ACC:{}*".format(",".join(formatted))
        return ""

    @property
    def _amount(self):
        if self.amount is not None:
            return f"AM:{self.amount:.2f}*"
        return ""

    @property
    def _due_date(self):
        if self.due_date is not None:
            str_part = "DT:{}*"
            if isinstance(self.due_date, datetime):
                return str_part.format(self.due_date.date().isoformat()).replace("-", "")
            if isinstance(self.due_date, date):
                return str_part.format(self.due_date.isoformat()).replace("-", "")
            return str_part.format(self.due_date)
        return ""

    def _format_item_string(self, item, name):
        if item is not None and item != "":
            return f"{name}:{item}*"
        return ""

    def get_text(self):
        return "SPD*1.0*{ACC}{ALTACC}{AM}{CC}{RF}{RN}{DT}{PT}{MSG}{NT}{NTA}{XPER}{XVS}{XSS}{XKS}{XID}{XURL}".format(
            ACC=self._account,
            ALTACC=self._alternate_accounts,
            AM=self._amount,
            CC=self._format_item_string(self.currency, "CC"),
            RF=self._format_item_string(self.reference, "RF"),
            RN=self._format_item_string(self.recipient_name, "RN"),
            DT=self._due_date,
            PT=self._format_item_string(self.payment_type, "PT"),
            MSG=self._format_item_string(self.message, "MSG"),
            NT=self._format_item_string(self.notification_type, "NT"),
            NTA=self._format_item_string(self.notification_address, "NTA"),
            XPER=self._format_item_string(self.x_per, "X-PER"),
            XVS=self._format_item_string(self.x_vs, "X-VS"),
            XSS=self._format_item_string(self.x_ss, "X-SS"),
            XKS=self._format_item_string(self.x_ks, "X-KS"),
            XID=self._format_item_string(self.x_id, "X-ID"),
            XURL=self._format_item_string(self.x_url, "X-URL"),
        ).rstrip("*")


def __getattr__(name):
    if name == "QRPlatbaGenerator":
        import warnings

        warnings.warn(
            "Importing QRPlatbaGenerator from qrplatba.spayd is deprecated. "
            "Use 'from qrplatba import QRPlatbaGenerator' instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        from qrplatba.generator import QRPlatbaGenerator

        return QRPlatbaGenerator
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



import os
from decimal import Decimal
from pathlib import Path
from typing import NamedTuple

from qrcode.compat.etree import ET
from qrcode.image import svg

_FONT_DIR = Path(__file__).parent / "fonts"
_INTER_BOLD = str(_FONT_DIR / "Inter-Bold.ttf")


class ScaledSizes(NamedTuple):
    inside_border: Decimal
    outside_border: Decimal
    width: Decimal
    line_size: Decimal
    ratio: Decimal


class QRPlatbaSVGImage(svg.SvgPathImage):
    """
    QR Platba SVG image generator.

    Inner padding is created according to specification (http://qr-platba.cz/pro-vyvojare/specifikace-formatu/),
    text size is computed to achieve width of 16 QR points.
    """

    QR_TEXT_STYLE = "font-size:{size}px;font-weight:bold;fill:#000000;font-family:Inter,Arial,Helvetica,sans-serif;"
    FONT_SIZE = Decimal("3.5")
    FONT_HEIGHT = Decimal("10")

    LINE_SIZE = Decimal("0.5")
    INSIDE_BORDER = 4

    BOTTOM_LINE_SEGMENTS = (2, 22)

    def __init__(self, border, width, box_size, *args, **kwargs):
        self.outside_border = border
        border += self.INSIDE_BORDER + self.LINE_SIZE  # outside border + inside border + line size

        super().__init__(border, width, box_size, *args, **kwargs)

    def _get_scaled_sizes(self):
        """Computes sizes of the QR code and QR text according to the scale ratio"""
        scale_ratio = self.units(self.box_size, text=False)

        def strip_zeros(value):
            return Decimal(str(value).rstrip("0").rstrip(".")) if "." in str(value) else value

        return ScaledSizes(
            inside_border=strip_zeros(self.INSIDE_BORDER * scale_ratio),
            outside_border=strip_zeros(self.outside_border * scale_ratio),
            width=strip_zeros(self.width * scale_ratio),
            line_size=strip_zeros(self.LINE_SIZE * scale_ratio),
            ratio=scale_ratio,
        )

    def make_border(self):
        """Creates black thin border around QR code"""
        scaled = self._get_scaled_sizes()

        def sizes(ob, ib, wd, ln):  # size helper
            return ob * scaled.outside_border + ib * scaled.inside_border + wd * scaled.width + ln * scaled.line_size

        horizontal_line = "M{x0},{y0}h{length}v{width}h-{length}z"
        vertical_line = "M{x0},{y0}v{length}h{width}v-{length}z"

        def get_subpaths():
            # top line
            yield horizontal_line.format(
                x0=scaled.outside_border,
                y0=scaled.outside_border,
                length=sizes(0, 2, 1, 2),
                width=scaled.line_size,
            )

            b_first, b_second = self.BOTTOM_LINE_SEGMENTS

            # bottom line - first segment
            yield horizontal_line.format(
                x0=scaled.outside_border,
                y0=sizes(1, 2, 1, 1),
                length=b_first * scaled.ratio,
                width=scaled.line_size,
            )

            # bottom line - second segment
            yield horizontal_line.format(
                x0=scaled.outside_border + b_second * scaled.ratio,
                y0=sizes(1, 2, 1, 1),
                length=sizes(0, 2, 1, 2) - b_second * scaled.ratio,
                width=scaled.line_size,
            )

            # left line
            yield vertical_line.format(
                x0=scaled.outside_border,
                y0=scaled.outside_border + scaled.line_size,
                length=scaled.width + 2 * scaled.inside_border,
                width=scaled.line_size,
            )

            # right line
            yield vertical_line.format(
                x0=sizes(1, 2, 1, 1),
                y0=sizes(1, 0, 0, 1),
                length=sizes(0, 2, 1, 0),
                width=scaled.line_size,
            )

        subpaths = " ".join(get_subpaths())
        return ET.Element("path", d=subpaths, id="qrplatba-border", **self.QR_PATH_STYLE)

    def make_text(self):
        """Creates "QR platba" text element"""
        scaled = self._get_scaled_sizes()
        text_style = self.QR_TEXT_STYLE.format(size=(self.FONT_SIZE * scaled.ratio).quantize(Decimal("0.01")))

        x_pos = str(scaled.outside_border + scaled.line_size + 3 * scaled.ratio)
        y_pos = str(
            scaled.outside_border
            + scaled.line_size
            + 2 * scaled.inside_border
            + scaled.width
            + (self.FONT_HEIGHT / 5) * scaled.ratio
        )

        text_el = ET.Element("text", style=text_style, x=x_pos, y=y_pos, id="qrplatba-text")
        text_el.text = "QR platba"

        return text_el

    def _svg(self, viewBox=None, **kwargs):
        scaled = self._get_scaled_sizes()
        h_pixels = self.pixel_size + (self.FONT_HEIGHT * scaled.ratio)

        box = f"0 0 {self.units(self.pixel_size, text=False)} {self.units(h_pixels, text=False)}"
        svg_el = super()._svg(viewBox=box, **kwargs)
        svg_el.append(self.make_border())
        svg_el.append(self.make_text())

        # update size of the SVG element
        svg_el.attrib["height"] = str(self.units(h_pixels))

        return svg_el

    def save(self, stream, kind=None, *, output_format=None, zoom=None, resvg_kwargs=None):
        if output_format is None:
            output_format = kind
        if output_format is None or output_format.upper() == "SVG":
            return super().save(stream, kind=kind)

        if output_format.upper() != "PNG":
            raise ValueError(f"Unsupported format: {output_format}")

        self._save_png(stream, zoom=zoom, resvg_kwargs=resvg_kwargs)

    def _save_png(self, stream, *, zoom=None, resvg_kwargs=None):
        try:
            import resvg_py
        except ImportError:
            raise ImportError(
                "PNG support requires the 'resvg-py' package. Install it with: pip install qrplatba[png]"
            ) from None

        if resvg_kwargs is None:
            resvg_kwargs = {"zoom": zoom} if zoom is not None else {}
        else:
            resvg_kwargs = {k: v for k, v in resvg_kwargs.items() if k not in ("svg_string", "svg_path")}
            if zoom is not None:
                resvg_kwargs["zoom"] = zoom

        # SVG uses mm units; 300 DPI produces print-quality output with readable text
        resvg_kwargs.setdefault("dpi", 300)
        resvg_kwargs.setdefault("shape_rendering", "crisp_edges")

        if "font_files" not in resvg_kwargs and "skip_system_fonts" not in resvg_kwargs:
            resvg_kwargs["font_files"] = [_INTER_BOLD]
            resvg_kwargs["skip_system_fonts"] = True

        svg_string = self.to_string(encoding="unicode")
        png_bytes = resvg_py.svg_to_bytes(svg_string=svg_string, **resvg_kwargs)

        if isinstance(stream, (str, bytes, os.PathLike)):
            with open(stream, "wb") as f:
                f.write(png_bytes)
        else:
            stream.write(png_bytes)



import qrcode

from qrplatba.spayd import SpaydGenerator
from qrplatba.svg import QRPlatbaSVGImage


class QRPlatbaGenerator(SpaydGenerator):
    """QR Platba generator -- creates SPAYD QR code images."""

    def make_image(self, border=2, box_size=10, error_correction=qrcode.constants.ERROR_CORRECT_M):
        qr = qrcode.QRCode(
            version=None,
            error_correction=error_correction,
            image_factory=QRPlatbaSVGImage,
            border=border,
            box_size=box_size,
        )
        qr.add_data(self.get_text())
        qr.make(fit=True)
        return qr.make_image()
